Package: MSstatsBig
Type: Package
Title: MSstats Preprocessing for Larger than Memory Data
Version: 1.11.1
Authors@R: c(
  person("Anthony", "Wu", email = "wu.anthon@northeastern.edu", role = c("aut", "cre")),
  person("Mateusz", "Staniak", email = "mtst@mstaniak.pl", role = "aut"),
  person("Devon", "Kohler", email = "kohler.d@northeastern.edu", role = "aut")) 
Description: MSstats package provide tools for preprocessing, summarization and
    differential analysis of mass spectrometry (MS) proteomics data. Recently, 
    some MS protocols enable acquisition of data sets that result in larger than
    memory quantitative data. MSstats functions are not able to process such data.
    MSstatsBig package provides additional converter functions that enable 
    processing larger than memory data sets.
License: Artistic-2.0
Encoding: UTF-8
Imports: arrow, data.table, DBI, dplyr, MSstats, MSstatsConvert, readr,
        sparklyr, utils, stats
Suggests: testthat, mockery, knitr, rmarkdown
VignetteBuilder: knitr
biocViews: MassSpectrometry, Proteomics, Software
Config/roxygen2/version: 8.0.0
git_url: https://git.bioconductor.org/packages/MSstatsBig
git_branch: devel
git_last_commit: 1527b08
git_last_commit_date: 2026-08-07
Repository: Bioconductor 3.24
Date/Publication: 2026-08-09
NeedsCompilation: no
Packaged: 2026-08-09 23:04:53 UTC; biocbuild
Author: Anthony Wu [aut, cre],
  Mateusz Staniak [aut],
  Devon Kohler [aut]
Maintainer: Anthony Wu <wu.anthon@northeastern.edu>

# Generated by roxygen2: do not edit by hand

export(MSstatsAddAnnotationBig)
export(MSstatsPreprocessBig)
export(bigDIANNtoMSstatsFormat)
export(bigFragPipetoMSstatsFormat)
export(bigSpectronauttoMSstatsFormat)
importFrom(MSstats,dataProcess)
importFrom(MSstats,groupComparison)
importFrom(MSstatsConvert,MSstatsClean)
importFrom(MSstatsConvert,MSstatsImport)
importFrom(MSstatsConvert,MSstatsMakeAnnotation)
importFrom(data.table,":=")
importFrom(data.table,.SD)
importFrom(data.table,setDT)
importFrom(data.table,setnames)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,sessionInfo)

# MSstatsBig

<!-- badges: start -->
[![Bioconductor Release Build](https://bioconductor.org/shields/build/release/bioc/MSstatsBig.svg)](https://bioconductor.org/checkResults/release/bioc-LATEST/MSstatsBig/)
[![Codecov test coverage](https://codecov.io/gh/Vitek-Lab/MSstatsBig/branch/devel/graph/badge.svg)](https://codecov.io/gh/Vitek-Lab/MSstatsBig/branch/devel)
[![Bioconductor Downloads Rank](https://bioconductor.org/shields/downloads/release/MSstatsBig.svg)](https://bioconductor.org/packages/stats/bioc/MSstatsBig/)
[![Years in Bioconductor](https://bioconductor.org/shields/years-in-bioc/MSstatsBig.svg)](https://bioconductor.org/packages/release/bioc/html/MSstatsBig.html#since)
[![License: Artistic-2.0](https://img.shields.io/badge/license-Artistic--2.0-blue.svg)](https://opensource.org/licenses/Artistic-2.0)
<!-- badges: end -->

MSstatsBig is an R/Bioconductor package for preprocessing mass spectrometry-based
proteomics experiments that are larger than a standard computer's memory. Modern
acquisition protocols — most often data-independent acquisition (DIA) with a
large number of MS runs — can produce quantitative reports that the standard
MSstats converters cannot load into RAM. MSstatsBig provides drop-in replacement
converters that filter each tool's PSM report down to the data required for
differential analysis using out-of-memory backends (Arrow / Spark), so the
result fits in memory. Once a converter has run, the standard MSstats workflow
(`dataProcess()`, `groupComparison()`) can be followed as usual.

MSstatsBig is part of the [MSstats](https://github.com/Vitek-Lab/MSstats)
family of packages, developed and maintained by the
[Vitek Lab](https://olga-vitek-lab.khoury.northeastern.edu/) at Northeastern
University. The package and its documentation are also available at
[msstats.org](http://msstats.org).

## Installation

```r
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

BiocManager::install("MSstatsBig")
```

The development version can be installed directly from this repository:

```r
remotes::install_github("Vitek-Lab/MSstatsBig")
```

## Quick Start

```r
library(MSstatsBig)

# Example FragPipe msstats.csv output bundled with the package
input <- system.file("extdata", "fgexample.csv", package = "MSstatsBig")

# Convert out-of-memory using the Arrow backend; writes to output_file.csv
converted <- bigFragPipetoMSstatsFormat(input,
                                        "output_file.csv",
                                        backend = "arrow",
                                        max_feature_count = 20)

# Collect the Arrow object into an in-memory data.frame
converted <- as.data.frame(dplyr::collect(converted))

# Continue with the standard MSstats workflow
library(MSstats)
summarized <- dataProcess(converted, use_log_file = FALSE)
```

Users can also feed data into the underlying `MSstatsPreprocessBig()` function
directly — either from a tool's native export format or by converting raw data
chunk by chunk — to support other tools. See the workflow
vignette for details.

## Supported Converters

MSstatsBig provides larger-than-memory replacements for the standard MSstats
converters. After a converter runs, its output is used with
[MSstats](https://github.com/Vitek-Lab/MSstats):

| Search tool / format | Converter function |
| --- | --- |
| FragPipe | `bigFragPipetoMSstatsFormat()` |
| Spectronaut | `bigSpectronauttoMSstatsFormat()` |
| DIA-NN | `bigDIANNtoMSstatsFormat()` |
| Generic MSstats-format data | `MSstatsPreprocessBig()` |

The `MSstatsAddAnnotationBig()` helper attaches experimental-design annotation
to converted data. See the [MSstatsBig workflow vignette](vignettes/MSstatsBig_Workflow.Rmd)
for the required input files and options for each converter.

## Documentation

- [MSstatsBig workflow](vignettes/MSstatsBig_Workflow.Rmd) — full worked example from raw data to results
- [Official website: msstats.org](http://msstats.org)
- [Bioconductor package page and reference manual](https://bioconductor.org/packages/MSstatsBig)

## Getting Help / Reporting Bugs

- **Questions about usage, statistical methods, or troubleshooting:** please
  post to the [MSstats Google Group](https://groups.google.com/forum/#!forum/msstats).
  This is monitored by the development team and searchable, so it's the fastest
  way to get help and to see if your question has already been answered.
- **Bug reports and feature requests for this repository:** please open a
  [GitHub issue](https://github.com/Vitek-Lab/MSstatsBig/issues).

## References

MSstatsBig prepares data for the standard MSstats workflow. If you use it, please
cite the MSstats publication(s):

1. Kohler D, Staniak M, Tsai TH, Huang T, Shulman N, Bernhardt OM, MacLean BX,
   Nesvizhskii AI, Reiter L, Sabido E, Choi M, Vitek O. **MSstats Version 4.0:
   Statistical Analyses of Quantitative Mass Spectrometry-Based Proteomic
   Experiments with Chromatography-Based Quantification at Scale.**
   *J Proteome Res*. 2023;22(5):1466-1482.
   [DOI: 10.1021/acs.jproteome.2c00834](https://doi.org/10.1021/acs.jproteome.2c00834)
2. Kohler D, Vitek O, et al. **An MSstats workflow for detecting differentially
   abundant proteins in large-scale data-independent acquisition mass
   spectrometry experiments with FragPipe processing.** *Nat Protoc*.
   2024;19:2915-2938.
   [DOI: 10.1038/s41596-024-01000-3](https://doi.org/10.1038/s41596-024-01000-3)

## Funding

MSstats development has been supported by the Chan Zuckerberg Initiative's
[Essential Open Source Software for Science](https://chanzuckerberg.com/eoss/proposals/).

## License

MSstatsBig is released under the [Artistic-2.0](https://opensource.org/licenses/Artistic-2.0) license.

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE)

## ----load_package-------------------------------------------------------------
library(MSstatsBig)

## ----example_data-------------------------------------------------------------
head(read.csv(system.file("extdata", "fgexample.csv", package = "MSstatsBig")))

## ----conv_example-------------------------------------------------------------
setwd(tempdir())

converted_data = bigFragPipetoMSstatsFormat(
  system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend="arrow",
  max_feature_count = 20)

# The returned arrow object needs to be collected for the remaining workflow
converted_data = as.data.frame(dplyr::collect(converted_data))

## ----msstats_workflow---------------------------------------------------------

library(MSstats)

# converted_data = read.csv("output_file.csv")
summarized_data = dataProcess(converted_data,
                              use_log_file = FALSE)

# Build contrast matrix
comparison = matrix(c(-1, 1),
    nrow=1, byrow=TRUE)
row.names(comparison) <- c("T-NAT")
colnames(comparison) <- c("NAT", "T")

model_results = groupComparison(contrast.matrix = comparison, 
                                data = summarized_data,
                                use_log_file = FALSE)



## -----------------------------------------------------------------------------
sessionInfo()

#' @inheritParams MSstatsPreprocessBig
#' @keywords internal
MSstatsPreprocessBigSparklyr <- function(connection, input_file, output_file_name,
                                         max_feature_count = 20,
                                         filter_unique_peptides = FALSE,
                                         aggregate_psms = FALSE,
                                         filter_few_obs = FALSE,
                                         remove_annotation = FALSE) {
  tbl_name <- getTblName(remove_annotation)
  
  sparklyr::spark_read_csv(connection, "mstinput", path = input_file,
                           header = TRUE, delimiter = ",",
                           memory = FALSE, overwrite = TRUE,
                           repartition = 100)
  
  if (remove_annotation) {
    dplyr::tbl(connection, "mstinput") %>%
      dplyr::select(-Condition, -BioReplicate) %>%
      sparklyr::spark_write_table("mstinputnoannot", mode = "overwrite")
    DBI::dbRemoveTable(connection, "mstinput")
  }
  
  dplyr::tbl(connection, "mstinputnoannot") %>%
    dplyr::group_by(PeptideSequence) %>%
    dplyr::summarize(NumProteins = n_distinct(ProteinName)) %>%
    dplyr::filter(NumProteins > 1) %>%
    sparklyr::spark_write_table("sharedpeptides", mode = "overwrite")
  sparklyr::sdf_repartition(tbl(connection, "mstinputnoannot"), partition_by = "ProteinName")
  
  dplyr::tbl(connection, "mstinputnoannot") %>%
    dplyr::group_by(ProteinName, PeptideSequence,
                    PrecursorCharge, FragmentIon, ProductCharge, IsotopeLabelType,
                    Run) %>%
    dplyr::summarize(Intensity = max(Intensity)) %>%
    sparklyr::spark_write_table("mstinputagg", mode = "overwrite")
  DBI::dbRemoveTable(connection, "mstinputnoannot")
  
  dplyr::tbl(connection, "mstinputagg") %>%
    dplyr::group_by(ProteinName, PeptideSequence, PrecursorCharge, FragmentIon, ProductCharge, IsotopeLabelType) %>%
    dplyr::summarize(NumObs = sum(as.numeric(!is.na(Intensity) & Intensity > 1))) %>%
    dplyr::filter(NumObs <= 2) %>%
    dplyr::select(-NumObs) %>%
    sparklyr::spark_write_table(name = "less2obs", mode = "overwrite")
  
  dplyr::anti_join(
    dplyr::tbl(connection, "mstinputagg"),
    tbl(connection, "less2obs"),
    by <- c("ProteinName",
            "PeptideSequence", "PrecursorCharge", "FragmentIon",
            "ProductCharge", "IsotopeLabelType")
  ) %>%
    sparklyr::spark_write_table(name = "mstinputfilt", mode = "overwrite")
  DBI::dbRemoveTable(connection, "mstinputagg")
  DBI::dbRemoveTable(connection, "less2obs")
  
  dplyr::tbl(connection, "mstinputfilt") %>%
    dplyr::group_by(ProteinName, PeptideSequence, PrecursorCharge, FragmentIon, ProductCharge, IsotopeLabelType) %>%
    dplyr::summarize(Intensity = mean(Intensity)) %>%
    dplyr::ungroup() %>%
    sparklyr::spark_write_table("fortop50", mode = "overwrite", partition_by = "ProteinName")
  
  
  dplyr::tbl(connection, "fortop50") %>%
    dplyr::group_by(ProteinName, IsotopeLabelType) %>%
    dplyr::mutate(Rank = min_rank(-Intensity)) %>%
    dplyr::filter(Rank <= max_feature_count) %>%
    dplyr::select(-Rank, -Intensity) %>%
    sparklyr::spark_write_table("top50", mode = "overwrite", partition_by = "ProteinName")
  
  dplyr::inner_join(
    dplyr::tbl(connection, "mstinputfilt"),
    dplyr::tbl(connection, "top50"),
    by <- c("ProteinName", "IsotopeLabelType", "PeptideSequence", "PrecursorCharge",
            "FragmentIon", "ProductCharge")) %>%
    sparklyr::spark_write_table("mstprocessed", mode = "overwrite")
  DBI::dbRemoveTable(connection, "top50")
  DBI::dbRemoveTable(connection, "fortop50")
  DBI::dbRemoveTable(connection, "mstinputfilt")
  
  sparklyr::spark_write_csv(sparklyr::sdf_repartition(dplyr::tbl(connection, "mstprocessed"), 1),
                            output_file_name, mode = "overwrite")
  dplyr::tbl(connection, "mstprocessed")
}


#' @inheritParams MSstatsPreprocessBig
#' @keywords internal
MSstatsPreprocessBigArrow <- function(input_file,
                                      output_file_name,
                                      max_feature_count = 20,
                                      filter_unique_peptides = FALSE,
                                      aggregate_psms = FALSE,
                                      filter_few_obs = FALSE,
                                      calculateAnomalyScores = FALSE, 
                                      anomalyModelFeatures = c()) {
  input <- arrow::open_dataset(input_file, format = "csv")
  
  # dynamically cast any columns Arrow inferred as 'null'
  null_columns <- names(input$schema)[vapply(input$schema$fields, function(f) f$type$ToString() == "null", logical(1))]
  if (length(null_columns) > 0) {
    for (col in null_columns) {
      input <- dplyr::mutate(input, !!rlang::sym(col) := as.numeric(!!rlang::sym(col)))
    }
  }

  if ("PrecursorMz" %in% input$schema$names) {
    input <- dplyr::mutate(input, PrecursorMz = as.numeric(PrecursorMz))
  }
  
  input <- dplyr::mutate(input,
                         Feature = paste(PeptideSequence, PrecursorCharge,
                                         FragmentIon, ProductCharge, sep = "_"))
  feature_counts <- dplyr::group_by(input, ProteinName, Feature)
  feature_counts <- dplyr::summarize(feature_counts,
                                     MeanAbundance = mean(Intensity,
                                                          na.rm = TRUE))
  feature_counts <- dplyr::collect(feature_counts)
  
  feature_counts <- dplyr::mutate(
    feature_counts,
    feature_rank = dplyr::min_rank(dplyr::desc(MeanAbundance)))
  
  feature_counts <- dplyr::filter(feature_counts,
                                  feature_rank <= max_feature_count)
  
  feature_counts <- dplyr::select(feature_counts, -MeanAbundance, -feature_rank)
  input <- dplyr::inner_join(input, feature_counts, 
                             by = c("ProteinName", "Feature"))
  input <- dplyr::select(input, -Feature)
  
  arrow::write_dataset(input, .prefixedPath("topN_", output_file_name), format = "csv")
  
  if (filter_unique_peptides) {
    pp_df <- dplyr::select(input, ProteinName, PeptideSequence)
    pp_df <- dplyr::group_by(pp_df, PeptideSequence)
    pp_df <- dplyr::summarize(pp_df, NumProteins = dplyr::n_distinct(ProteinName))
    pp_df <- dplyr::filter(pp_df, NumProteins > 1)
    pp_df <- dplyr::select(pp_df, -NumProteins)
    input <- dplyr::anti_join(input, pp_df, by = "PeptideSequence")
  }
  if (aggregate_psms) {
    group_cols <- c("ProteinName", "PeptideSequence", "PrecursorCharge",
                    "FragmentIon", "ProductCharge", "IsotopeLabelType", "Run",
                    "Condition", "BioReplicate")
    
    max_per_group <- input %>%
      group_by(across(all_of(group_cols))) %>%
      summarise(max_intensity = max(Intensity, na.rm = TRUE), 
                .groups = "drop")
    
    input <- input %>%
      inner_join(max_per_group, by = group_cols) %>%
      filter(Intensity == max_intensity) %>%
      select(-max_intensity)
  }
  
  if (filter_few_obs) {
    input <- dplyr::mutate(input,
                             Feature = paste(PeptideSequence, PrecursorCharge,
                                             FragmentIon, ProductCharge, sep = "_"))
    input <- dplyr::group_by(input, ProteinName, Feature)
    observation_counts <- dplyr::summarize(input,
                                           NumObs = sum(!is.na(Intensity) &
                                                          Intensity > 0))
    observation_counts <- dplyr::filter(observation_counts, NumObs <= 2)
    observation_counts <- dplyr::select(observation_counts, -NumObs)
    input <- dplyr::anti_join(input, observation_counts,
                              by = c("ProteinName", "Feature"))
    input <- dplyr::select(input, -Feature)
  }
  
  arrow::write_dataset(input, output_file_name, format = "csv")
  input
}

getTblName <- function(remove_annotation) {
  if (remove_annotation) {
    "mstinput"
  } else {
    "mstinputnoannot"
  }
}

#' Build an intermediate output path by prefixing only the basename.
#'
#' Naive `paste0(prefix, output_file_name)` corrupts paths that contain a
#' directory (`subdir/out.csv` → `topN_subdir/out.csv`,
#' `/tmp/out.csv` → `topN_/tmp/out.csv`). Splitting via dirname/basename keeps
#' the directory component intact so intermediate files land beside the final
#' output.
#'
#' @param prefix Character scalar prepended to the basename.
#' @param path  Output file path supplied by the caller.
#' @return Character scalar.
#' @keywords internal
.prefixedPath <- function(prefix, path) {
    file.path(dirname(path), paste0(prefix, basename(path)))
}

#' Write chunk to file
#'
#' @param input Data frame
#' @param output_path Path to output file
#' @param pos Chunk position
#' @return NULL
#' @keywords internal
.writeChunkToFile <- function(input, output_path, pos) {
    # Write to file
    if (!is.null(pos)) {
        if (pos == 1) {
            readr::write_csv(input, file = output_path, append = FALSE)
        } else {
            readr::write_csv(input, file = output_path, append = TRUE)
        }
    }
}

#' @importFrom data.table := .SD setDT setnames
#' @importFrom stats setNames
NULL

#' @keywords internal
reduceBigSpectronaut <- function(input_file, output_path,
                                 intensity="F.NormalizedPeakArea",
                                 filter_by_excluded = FALSE,
                                 filter_by_identified = FALSE,
                                 filter_by_qvalue = TRUE,
                                 qvalue_cutoff = 0.01,
                                 calculateAnomalyScores=FALSE,
                                 anomalyModelFeatures=c(),
                                 block_size = 16L * 1024L * 1024L) {
  block_size <- as.integer(block_size)
  stopifnot(length(block_size) == 1L, !is.na(block_size), block_size > 0L)

  if (grepl("csv", input_file)) {
    delim <- ","
  } else if (grepl("tsv|xls", input_file)) {
    delim <- "\t"
  } else {
    delim <- ";"
  }

  needed_cols <- c("R.FileName", "R.Condition", "R.Replicate",
                   "PG.ProteinAccessions", "EG.ModifiedSequence",
                   "FG.LabeledSequence", "FG.Charge",
                   "F.FrgIon", "F.Charge",
                   "EG.Identified", "F.ExcludedFromQuantification",
                   "F.FrgLossType", "PG.Qvalue", "EG.Qvalue",
                   intensity)
  if (calculateAnomalyScores) {
    needed_cols <- c(needed_cols, anomalyModelFeatures)
  }

  parse_opts   <- arrow::CsvParseOptions$create(delimiter = delim)
  convert_opts <- arrow::CsvConvertOptions$create()
  read_opts    <- arrow::CsvReadOptions$create(block_size = block_size)

  ds <- arrow::open_dataset(
    input_file,
    format          = "csv",
    parse_options   = parse_opts,
    convert_options = convert_opts,
    read_options    = read_opts
  )

  present_cols <- intersect(needed_cols, names(ds))
  scanner <- arrow::Scanner$create(ds, projection = present_cols)
  reader <- scanner$ToRecordBatchReader()

  t_start   <- Sys.time()
  pos       <- 1L
  batch_idx <- 0L
  repeat {
    batch <- reader$read_next_batch()
    if (is.null(batch)) break
    chunk_df <- as.data.frame(batch)
    cleanSpectronautChunk(chunk_df,
                          output_path,
                          intensity,
                          filter_by_excluded,
                          filter_by_identified,
                          filter_by_qvalue,
                          qvalue_cutoff,
                          pos,
                          calculateAnomalyScores,
                          anomalyModelFeatures)
    pos       <- pos + nrow(chunk_df)
    batch_idx <- batch_idx + 1L

    if (batch_idx %% 1000L == 0L) {
      elapsed <- as.numeric(Sys.time() - t_start, units = "secs")
      rate    <- (pos - 1L) / elapsed
      message(sprintf(
        "[reduceBigSpectronaut] %d batches | %s rows | %.1fk rows/s | %.0fs elapsed",
        batch_idx,
        format(pos - 1L, big.mark = ","),
        rate / 1000,
        elapsed))
    }

    rm(batch, chunk_df)
  }

  if (batch_idx %% 1000L != 0L) {
    elapsed <- as.numeric(Sys.time() - t_start, units = "secs")
    rate    <- (pos - 1L) / elapsed
    message(sprintf(
      "[reduceBigSpectronaut] done: %d batches | %s rows | %.1fk rows/s | %.0fs elapsed",
      batch_idx,
      format(pos - 1L, big.mark = ","),
      rate / 1000,
      elapsed))
  }
}

#' @keywords internal
cleanSpectronautChunk = function(input, output_path,
                                 intensity="F.NormalizedPeakArea",
                                 filter_by_excluded = FALSE,
                                 filter_by_identified = FALSE,
                                 filter_by_qvalue = TRUE,
                                 qvalue_cutoff = 0.01,
                                 pos = NULL,
                                 calculateAnomalyScores=FALSE,
                                 anomalyModelFeatures=c()) {
  data.table::setDT(input)

  all_cols <- c("R.FileName", "R.Condition", "R.Replicate",
                "PG.ProteinAccessions", "EG.ModifiedSequence", "FG.LabeledSequence",
                "FG.Charge", "F.FrgIon", "F.Charge",
                "EG.Identified", "F.ExcludedFromQuantification", "F.FrgLossType",
                "PG.Qvalue", "EG.Qvalue", intensity)
  new_names <- c("Run", "Condition", "BioReplicate", "ProteinName",
                 "PeptideSequence", "LabeledSequence", "PrecursorCharge", "FragmentIon",
                 "ProductCharge", "Identified", "Excluded",
                 "FFrgLossType", "PGQvalue", "EGQvalue",
                 "Intensity")
  if (calculateAnomalyScores) {
    all_cols  <- c(all_cols, anomalyModelFeatures)
    new_names <- c(new_names, MSstatsConvert:::.standardizeColnames(anomalyModelFeatures))
  }

  present_orig <- intersect(all_cols, colnames(input))
  if (length(present_orig) == 0L) {
    stop(sprintf(
      paste0("cleanSpectronautChunk: none of the expected Spectronaut ",
             "columns were found in the input batch. ",
             "Expected any of: %s. Found: %s. ",
             "Check that the file is comma-delimited and that the ",
             "Spectronaut export uses the standard column names."),
      paste(all_cols, collapse = ", "),
      paste(colnames(input), collapse = ", ")))
  }
  input <- input[, present_orig, with = FALSE]

  data.table::setnames(input, MSstatsConvert:::.standardizeColnames(colnames(input)))
  spectronaut_to_msstats_col_mapping <- stats::setNames(new_names,
                                    MSstatsConvert:::.standardizeColnames(all_cols))
  data.table::setnames(input,
                       old = names(spectronaut_to_msstats_col_mapping),
                       new = unname(spectronaut_to_msstats_col_mapping),
                       skip_absent = TRUE)

  input[, Intensity := as.numeric(Intensity)]

  if (is.character(input[["Excluded"]])) {
    input[, Excluded := Excluded == "True"]
  }
  if ("Identified" %in% colnames(input) && is.character(input[["Identified"]])) {
    input[, Identified := Identified == "True"]
  }

  msstats_to_spectronaut_col_mapping <- stats::setNames(all_cols, new_names)
  require_filter_cols <- function(cols, filter_name) {
    missing <- setdiff(cols, colnames(input))
    if (length(missing) > 0L) {
      stop(sprintf(
        paste0("cleanSpectronautChunk: %s needs Spectronaut column(s) %s, ",
               "which were not found in the input export. Found: %s."),
        filter_name,
        paste(msstats_to_spectronaut_col_mapping[missing], collapse = ", "),
        paste(colnames(input), collapse = ", ")))
    }
  }

  if (filter_by_excluded) {
    require_filter_cols("Excluded", "filter_by_excluded")
    input[Excluded == TRUE, Intensity := NA_real_]
  }
  if (filter_by_identified) {
    require_filter_cols("Identified", "filter_by_identified")
    input[Identified == FALSE, Intensity := NA_real_]
  }
  if (filter_by_qvalue) {
    require_filter_cols(c("EGQvalue", "PGQvalue"), "filter_by_qvalue")
    input[is.na(EGQvalue) | EGQvalue >= qvalue_cutoff, Intensity := NA_real_]
    input[is.na(PGQvalue) | PGQvalue >= qvalue_cutoff, Intensity := NA_real_]
  }

  require_filter_cols("FFrgLossType", "the noloss fragment-loss filter")
  input <- input[FFrgLossType == "noloss"]
  input[, IsotopeLabelType := "L"]

  select_cols <- c("ProteinName", "PeptideSequence", "PrecursorCharge", "FragmentIon",
                   "ProductCharge", "IsotopeLabelType", "Run", "BioReplicate", "Condition",
                   "Intensity")
  if (calculateAnomalyScores) {
    select_cols <- c(select_cols,
                     MSstatsConvert:::.standardizeColnames(anomalyModelFeatures))
  }

  input <- input[, select_cols, with = FALSE]
  .writeChunkToFile(input, output_path, pos)
  NULL
}

#' Read and clean a large DIANN file in chunks
#' 
#' @param input_file Path to the input DIANN file
#' @param output_path Path to the output CSV file
#' @param MBR Boolean, whether MBR was used
#' @param quantificationColumn Name of the column containing intensity values
#' @param global_qvalue_cutoff Global Q-value cutoff
#' @param qvalue_cutoff Q-value cutoff
#' @param pg_qvalue_cutoff Protein group Q-value cutoff
#' @param calculateAnomalyScores Boolean for MSstats+ Model
#' @param anomalyModelFeatures Character vector of features to use for MSstats+ Model
#' @param annotation Annotation file or data frame
#' @return NULL. Writes to file.
#' @keywords internal
reduceBigDIANN <- function(input_file, output_path, MBR = TRUE,
                           quantificationColumn = "FragmentQuantCorrected",
                           global_qvalue_cutoff = 0.01,
                           qvalue_cutoff = 0.01,
                           pg_qvalue_cutoff = 0.01,
                           calculateAnomalyScores=FALSE,
                           anomalyModelFeatures=c(),
                           annotation = NULL) {
  # Per-chunk callback shared by both the parquet and delimited-text paths.
  # `pos` drives .writeChunkToFile: pos == 1 overwrites, pos > 1 appends.
  diann_chunk <- function(x, pos) cleanDIANNChunk(x, output_path, MBR,
                                                  quantificationColumn, pos,
                                                  global_qvalue_cutoff,
                                                  qvalue_cutoff,
                                                  pg_qvalue_cutoff,
                                                  calculateAnomalyScores,
                                                  anomalyModelFeatures,
                                                  annotation)

  # Parquet branch (DIANN 2.0+): stream record batches via arrow so the file
  # is never fully materialised. read_delim_chunked can't read parquet bytes.
  if (tolower(tools::file_ext(input_file)) == "parquet") {
    # Lazy handle to the parquet file — no data loaded yet.
    ds <- arrow::open_dataset(input_file, format = "parquet")
    # Scanner + RecordBatchReader yields one batch at a time on demand.
    # batch_size matches the delimited-text path's 1M-row chunks; row-group
    # boundaries in the parquet may cap individual batches below this.
    scanner <- arrow::Scanner$create(ds, batch_size = 1e6)
    reader <- scanner$ToRecordBatchReader()
    pos <- 1
    repeat {
      batch <- reader$read_next_batch()
      if (is.null(batch)) break  # exhausted
      # Materialise just this batch, run it through the shared cleaner.
      diann_chunk(as.data.frame(batch), pos)
      pos <- pos + 1
    }
    return(invisible(NULL))
  }

  # Delimited-text branch (DIANN 1.x TSV/CSV): sniff the delimiter from the
  # first line, defaulting to tab when nothing matches.
  first_line <- readLines(input_file, n = 1)
  if (grepl("\t", first_line)) {
    delim <- "\t"
  } else if (grepl(",", first_line)) {
    delim <- ","
  } else if (grepl(";", first_line)) {
    delim <- ";"
  } else {
    delim <- "\t"
  }

  # Stream the file in 1M-row chunks, invoking diann_chunk for each.
  readr::read_delim_chunked(input_file,
                            readr::DataFrameCallback$new(diann_chunk),
                            delim = delim,
                            chunk_size = 1e6)
}

#' Clean a single chunk of DIANN data
#' 
#' @param input Data frame chunk
#' @param output_path Path to output file
#' @param MBR Boolean, whether MBR was used
#' @param quantificationColumn Name of intensity column
#' @param pos Chunk position (1 for first chunk, >1 for subsequent)
#' @param global_qvalue_cutoff Global Q-value cutoff
#' @param qvalue_cutoff Q-value cutoff
#' @param pg_qvalue_cutoff Protein group Q-value cutoff
#' @param calculateAnomalyScores Boolean for MSstats+ Model
#' @param anomalyModelFeatures Character vector of features to use for MSstats+ Model
#' @param annotation Annotation file or data frame
#' @importFrom MSstatsConvert MSstatsImport MSstatsClean MSstatsMakeAnnotation
#' @return NULL
#' @keywords internal
cleanDIANNChunk = function(input, output_path, MBR, quantificationColumn, pos,
                           global_qvalue_cutoff = 0.01,
                           qvalue_cutoff = 0.01,
                           pg_qvalue_cutoff = 0.01,
                           calculateAnomalyScores=FALSE,
                           anomalyModelFeatures = c(),
                           annotation = NULL) {
    input = MSstatsImport(list(input = input),
                          "MSstats", "DIANN")
    input = MSstatsClean(
        input, 
        MBR = MBR, 
        quantificationColumn = quantificationColumn,
        global_qvalue_cutoff = global_qvalue_cutoff, 
        qvalue_cutoff = qvalue_cutoff,
        pg_qvalue_cutoff = pg_qvalue_cutoff,
        calculateAnomalyScores = calculateAnomalyScores,
        anomalyModelFeatures = anomalyModelFeatures
    )
    if (!is.element("IsotopeLabelType", colnames(input))) {
        input <- dplyr::mutate(input, IsotopeLabelType = "L")
    }
    #input = MSstatsMakeAnnotation(input, annotation)
    .writeChunkToFile(input, output_path, pos)
    NULL
}

library(testthat)

context("DIANN converter functions")

# Test for the internal cleanDIANNChunk function
test_that("cleanDIANNChunk processes data correctly", {
  output_file <- tempfile(fileext = ".csv")

  diann_chunk_data <- data.frame(
    Run = "run1",
    Protein.Names = "ProteinA",
    Stripped.Sequence = "PEPTIDE",
    Modified.Sequence = "PEPTIDE(mod)",
    Precursor.Charge = 2,
    Fragment.Quant.Corrected = "100;200",
    Q.Value = 0.005,
    Precursor.Mz = 400.5,
    Fragment.Info = "y7^1/1;b3-H2O^1/1", # One valid, one to be filtered
    Lib.Q.Value = 0.001,
    Lib.PG.Q.Value = 0.001,
    stringsAsFactors = FALSE
  )

  # The function is not exported, so we use :::
  MSstatsBig:::cleanDIANNChunk(diann_chunk_data, output_file, MBR = TRUE,
                               quantificationColumn = "FragmentQuantCorrected", pos = 1)

  result <- read.csv(output_file)

  expect_equal(nrow(result), 1)
  expect_equal(result$ProteinName, "ProteinA")
  expect_equal(result$PeptideSequence, "PEPTIDE(mod)")
  expect_equal(result$Intensity, 100)
  expect_equal(result$FragmentIon, "y7^1/1")
  expect_true("PeptideSequence" %in% colnames(result))

  file.remove(output_file)
})

test_that("cleanDIANNChunk handles 'auto' quantification column correctly", {
  output_file <- tempfile(fileext = ".csv")

  # Data with wide format fragment quantification
  diann_chunk_wide <- data.frame(
    Run = "run1",
    Protein.Names = "ProteinA",
    Stripped.Sequence = "PEPTIDE",
    Modified.Sequence = "PEPTIDE",
    Precursor.Charge = 2,
    Fr1Quantity = 100,
    Fr2Quantity = 200,
    Q.Value = 0.005,
    Precursor.Mz = 400.5,
    Fragment.Info = "y1^1/1;y2^1/1",
    Lib.Q.Value = 0.001,
    Lib.PG.Q.Value = 0.001,
    stringsAsFactors = FALSE
  )

  MSstatsBig:::cleanDIANNChunk(diann_chunk_wide, output_file, MBR = TRUE,
                               quantificationColumn = "auto", pos = 1)

  result <- read.csv(output_file)

  expect_equal(nrow(result), 2)
  expect_equal(sort(result$Intensity), c(100, 200))
  expect_equal(sort(result$FragmentIon), c("y1^1/1", "y2^1/1"))

  file.remove(output_file)

  # Test error when columns are missing
  diann_chunk_missing <- diann_chunk_wide[, !grepl("Quantity", names(diann_chunk_wide))]
  expect_error(MSstatsBig:::cleanDIANNChunk(diann_chunk_missing, output_file, MBR = TRUE,
                                            quantificationColumn = "auto", pos = 1),
               "No fragment quantification columns found")
})

test_that("cleanDIANNChunk handles missing Fragment.Info by defaulting ProductCharge to 1", {
  output_file <- tempfile(fileext = ".csv")

  # Data with missing Fragment.Info (simulating it not being present)
  diann_chunk_missing <- data.frame(
    Run = "run1",
    Protein.Names = "ProteinA",
    Stripped.Sequence = "PEPTIDE",
    Modified.Sequence = "PEPTIDE",
    Precursor.Charge = 2,
    Fragment.Quant.Corrected = 100,
    Q.Value = 0.005,
    Precursor.Mz = 400.5,
    # Fragment.Info is missing
    Lib.Q.Value = 0.001,
    Lib.PG.Q.Value = 0.001,
    stringsAsFactors = FALSE
  )

  MSstatsBig:::cleanDIANNChunk(diann_chunk_missing, output_file, MBR = TRUE,
                               quantificationColumn = "FragmentQuantCorrected", pos = 1)

  result <- read.csv(output_file)

  expect_equal(nrow(result), 1)
  expect_equal(result$ProductCharge, 1)
  expect_equal(result$FragmentIon, "Frag1")

  file.remove(output_file)
})

# Test for the internal reduceBigDIANN function
test_that("reduceBigDIANN processes a file correctly", {
  input_file <- tempfile(fileext = ".csv")
  output_file <- tempfile(fileext = ".csv")

  diann_data <- data.frame(
    Run = c("run1", "run1"),
    Protein.Names = c("ProteinA", "ProteinB"),
    Stripped.Sequence = c("PEPTIDE_A", "PEPTIDE_B"),
    Modified.Sequence = c("PEPTIDE_A(mod)", "PEPTIDE_B"),
    Precursor.Charge = c(2, 3),
    Fragment.Quant.Corrected = c("100;200", "300"),
    Q.Value = c(0.005, 0.006),
    Precursor.Mz = c(400.5, 500.5),
    Fragment.Info = c("y7^1/1;b3-H2O^1/1", "y5^1/2"),
    Lib.Q.Value = c(0.001, 0.002),
    Lib.PG.Q.Value = c(0.001, 0.002),
    stringsAsFactors = FALSE
  )
  write.csv(diann_data, input_file, row.names = FALSE)

  MSstatsBig:::reduceBigDIANN(input_file, output_file, MBR = TRUE,
                              quantificationColumn = "FragmentQuantCorrected")

  result <- read.csv(output_file)
  expect_equal(nrow(result), 2)
  expect_equal(result$Intensity, c(100, 300))
  expect_equal(result$ProteinName, c("ProteinA", "ProteinB"))
  expect_equal(result$FragmentIon, c("y7^1/1", "y5^1/2"))

  file.remove(input_file)
  file.remove(output_file)
})

# End-to-end test for bigDIANNtoMSstatsFormat
test_that("bigDIANNtoMSstatsFormat works with arrow backend", {
  input_file <- tempfile(fileext = ".csv")
  output_file <- basename(tempfile(fileext = ".csv"))

  # 4 features for one protein. Feature selection should pick the top 2.
  diann_data <- rbind(
    data.frame(Run = c("r1", "r2"), Protein.Names = "P1", Stripped.Sequence = "PEPTIDE", Modified.Sequence = "PEPTIDE", Precursor.Charge = 2, Fragment.Quant.Corrected = c(1000, 1100), Q.Value = 0.001, Precursor.Mz = 500, Fragment.Info = "y1", Lib.Q.Value = 0.001, Lib.PG.Q.Value = 0.001),
    data.frame(Run = c("r1", "r2"), Protein.Names = "P1", Stripped.Sequence = "PEPTIDE", Modified.Sequence = "PEPTIDE", Precursor.Charge = 2, Fragment.Quant.Corrected = c(500, 600), Q.Value = 0.001, Precursor.Mz = 500, Fragment.Info = "y2", Lib.Q.Value = 0.001, Lib.PG.Q.Value = 0.001),
    data.frame(Run = c("r1", "r2"), Protein.Names = "P1", Stripped.Sequence = "PEPTIDE", Modified.Sequence = "PEPTIDE", Precursor.Charge = 2, Fragment.Quant.Corrected = c(100, 100), Q.Value = 0.001, Precursor.Mz = 500, Fragment.Info = "y3", Lib.Q.Value = 0.001, Lib.PG.Q.Value = 0.001),
    data.frame(Run = c("r1", "r2"), Protein.Names = "P1", Stripped.Sequence = "PEPTIDE", Modified.Sequence = "PEPTIDE", Precursor.Charge = 2, Fragment.Quant.Corrected = c(2000, 2100), Q.Value = 0.001, Precursor.Mz = 500, Fragment.Info = "y4", Lib.Q.Value = 0.001, Lib.PG.Q.Value = 0.001)
  )
  write.csv(diann_data, input_file, row.names = FALSE)
  converted <- bigDIANNtoMSstatsFormat(
    input_file = input_file,
    output_file_name = output_file,
    backend = "arrow",
    max_feature_count = 2
  )
  result <- dplyr::collect(converted)

  # Avg intensities: y1=1050, y2=550, y3=100, y4=2050.
  # Top 2 features are y4 and y1.
  expect_equal(nrow(result), 4) # 2 features * 2 runs
  expect_true(all(c("y1", "y4") %in% unique(result$FragmentIon)))
  expect_false(any(c("y2", "y3") %in% unique(result$FragmentIon)))

  # Cleanup — outputs may be directories when backend = "arrow"
  unlink(input_file, force = TRUE)
  unlink(output_file, recursive = TRUE, force = TRUE)
  unlink(paste0("reduce_output_", output_file), recursive = TRUE, force = TRUE)
  unlink(paste0("topN_", output_file), recursive = TRUE, force = TRUE)
})

test_that("bigDIANNtoMSstatsFormat works with annotation", {
  input_file <- tempfile(fileext = ".csv")
  output_file <- basename(tempfile(fileext = ".csv"))

  # Minimal DIANN data
  diann_data <- data.frame(
    Run = c("r1", "r2"), Protein.Names = "P1", Stripped.Sequence = "PEPTIDE", 
    Modified.Sequence = "PEPTIDE", Precursor.Charge = 2, 
    Fragment.Quant.Corrected = c(1000, 1100), Q.Value = 0.001, Precursor.Mz = 500, 
    Fragment.Info = "y1", Lib.Q.Value = 0.001, Lib.PG.Q.Value = 0.001
  )
  write.csv(diann_data, input_file, row.names = FALSE)
  
  # Annotation data
  annot <- data.frame(
    Run = c("r1", "r2"),
    Condition = c("Disease", "Healthy"),
    BioReplicate = c(1, 2)
  )
  
  converted <- bigDIANNtoMSstatsFormat(
    input_file = input_file,
    annotation = annot,
    output_file_name = output_file,
    backend = "arrow"
  )
  result <- dplyr::collect(converted)

  expect_true(all(c("Condition", "BioReplicate") %in% colnames(result)))
  expect_equal(result$Condition[result$Run == "r1"], "Disease")
  expect_equal(result$Condition[result$Run == "r2"], "Healthy")

  # Cleanup — outputs may be directories when backend = "arrow"
  unlink(input_file, force = TRUE)
  unlink(output_file, recursive = TRUE, force = TRUE)
  unlink(paste0("reduce_output_", output_file), recursive = TRUE, force = TRUE)
  unlink(paste0("topN_", output_file), recursive = TRUE, force = TRUE)
  unlink(paste0("cleaned_", output_file), recursive = TRUE, force = TRUE)
})

#' General converter for larger-than-memory csv files in MSstats format 10-column format
#'
#' @param input_file name of the input text file in 10-column MSstats format.
#' @param output_file_name name of an output file which will be saved after pre-processing
#' @param backend "arrow" or "sparklyr". Option "sparklyr" requires a spark installation
#' and connection to spark instance provided in the `connection` parameter.
#' @param max_feature_count maximum number of features per protein. Features will
#' be selected based on highest average intensity.
#' @param filter_unique_peptides If TRUE, shared peptides will be removed.
#' Please refer to the `Details` section for additional information.
#' @param aggregate_psms If TRUE, multiple measurements per PSM in a Run will
#' be aggregated (by taking maximum value). Please refer to the `Details` section for additional information.
#' @param filter_few_obs If TRUE, feature with less than 3 observations across runs will be removed.
#' Please refer to the `Details` section for additional information.
#' @param remove_annotation If TRUE, columns BioReplicate and Condition will be removed
#' to reduce output file size. These will need to be added manually later before
#' using dataProcess function. Only applicable to sparklyr backend.
#' @param calculateAnomalyScores If TRUE, will carry anomaly model features through pipeline
#' @param anomalyModelFeatures Character vector of column names to be carried through the pipeline
#' @param connection Connection to a spark instance created with the
#' `spark_connect` function from `sparklyr` package.
#'
#' @details Filtering and aggregation may be very time consuming and the ability
#' to perform them in a given R session depends on available memory, settings of
#' external packages, etc. Hence, all value of related parameters (`filter_unique_peptides`,
#' `aggregate_psms`, `filter_few_obs`) are set to FALSE by default and only feature
#' selection is performed, which saves both computation time and memory.
#' Appropriately configured spark backend provides the most consistent way to
#' perform these operations.
#'
#' @return either arrow object or sparklyr table that can be optionally collected
#' into memory by using dplyr::collect function.
#'
#' @examples
#' converted_data <- bigFragPipetoMSstatsFormat(
#'   system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
#'   "tencol_format.csv",
#'   backend="arrow")
#' procd <- MSstatsPreprocessBig("tencol_format.csv", "proc_out.csv", backend = "arrow")
#' head(dplyr::collect(procd))
#'
#' @export
#'
MSstatsPreprocessBig <-  function(input_file,
                                 output_file_name,
                                 backend,
                                 max_feature_count =  100,
                                 filter_unique_peptides =  FALSE,
                                 aggregate_psms =  FALSE,
                                 filter_few_obs =  FALSE,
                                 remove_annotation =  FALSE,
                                 calculateAnomalyScores = FALSE, 
                                 anomalyModelFeatures = c(),
                                 connection =  NULL) {
  if (backend == "arrow") {
    MSstatsPreprocessBigArrow(input_file,
                              output_file_name,
                              max_feature_count,
                              filter_unique_peptides,
                              aggregate_psms,
                              filter_few_obs,
                              calculateAnomalyScores, 
                              anomalyModelFeatures)
  } else if (backend == "sparklyr") {
    MSstatsPreprocessBigSparklyr(connection, input_file, output_file_name,
                                 max_feature_count, filter_unique_peptides,
                                 aggregate_psms, filter_few_obs,
                                 remove_annotation)
  } else {
    stop("backend not implemented")
  }
}

#' Convert out-of-memory FragPipe files to MSstats format.
#'
#' @inheritParams MSstatsPreprocessBig
#'
#' @export
#'
#' @examples
#' converted_data <- bigFragPipetoMSstatsFormat(
#'   system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
#'   "output_file.csv",
#'   backend = "arrow")
#' converted_data <- dplyr::collect(converted_data)
#' head(converted_data)
#'
#' @return either arrow object or sparklyr table that can be optionally collected
#' into memory by using dplyr::collect function.
#'
bigFragPipetoMSstatsFormat <-  function(input_file, output_file_name,
                                       backend,
                                       max_feature_count =  100,
                                       filter_unique_peptides =  FALSE,
                                       aggregate_psms =  FALSE,
                                       filter_few_obs =  FALSE,
                                       remove_annotation =  FALSE,
                                       connection =  NULL) {
  MSstatsPreprocessBig(
    input_file = input_file, 
    output_file_name = output_file_name,
    backend = backend, 
    max_feature_count = max_feature_count, 
    filter_unique_peptides = filter_unique_peptides,
    aggregate_psms = aggregate_psms, 
    filter_few_obs = filter_few_obs, 
    remove_annotation = remove_annotation,
    connection = connection)
}


#' Convert out-of-memory Spectronaut files to MSstats format.
#'
#' @inheritParams MSstatsPreprocessBig
#' @param annotation Optional data.frame with columns `Run`,
#'   `BioReplicate`, `Condition` (plus any additional annotation
#'   columns). If supplied, the converter merges it onto the output
#'   on `Run` and overrides any `Condition` / `BioReplicate` values
#'   that came from Spectronaut's `R.Condition` / `R.Replicate`
#'   columns. Required when the experimental design cannot be
#'   expressed in Spectronaut's own annotation — most notably paired
#'   designs, where `BioReplicate` must encode the pairing
#'   structure rather than per-sample IDs.
#' @param intensity Name of the intensity column to be used in Spectronaut
#' @param filter_by_excluded if TRUE, will filter by the `F.ExcludedFromQuantification` column.
#' @param filter_by_identified if TRUE, will filter by the `EG.Identified` column.
#' @param filter_by_qvalue if TRUE, will filter by EG.Qvalue and PG.Qvalue columns.
#' @param qvalue_cutoff cutoff which will be used for q-value filtering.
#' @param block_size Arrow CSV reader block size in bytes: the amount of raw
#'   CSV that is read and parsed at a time (roughly one record batch per block).
#'   Defaults to 16 MiB (`16L * 1024L * 1024L`). Two things matter when tuning it:
#'   \itemize{
#'     \item It must be larger than your single widest row. If you see
#'       `Invalid: straddling object straddles two block boundaries` on
#'       extra-wide exports, raise it (e.g. `64L * 1024L * 1024L`).
#'     \item Larger blocks reduce parsing overhead, but peak memory grows
#'       roughly linearly with `block_size`, so it is also the main lever on
#'       out-of-memory risk. Gains taper off quickly because per-batch time is
#'       dominated by row cleaning, not by block size.
#'   }
#'   To speed up processing without risking a crash, double the value
#'   (32 -> 64 -> 128 MiB) while watching memory use, and stop once throughput
#'   stops improving. Keep it comfortably below free RAM -- and below free RAM
#'   divided by the number of jobs if you run several conversions in parallel.
#'
#' @details
#' `bigSpectronauttoMSstatsFormat` consumes the following Spectronaut export
#' columns.
#'
#' Always required:
#' \itemize{
#'   \item `R.FileName`, `R.Condition`, `R.Replicate` -> Run, Condition, BioReplicate
#'   \item `PG.ProteinAccessions` -> ProteinName
#'   \item `EG.ModifiedSequence` -> PeptideSequence
#'   \item `FG.Charge` -> PrecursorCharge; `F.FrgIon` -> FragmentIon; `F.Charge` -> ProductCharge
#'   \item the intensity column (`F.NormalizedPeakArea` by default; see `intensity`) -> Intensity
#'   \item `F.FrgLossType` -> used by the always-on `noloss` fragment filter
#' }
#'
#'
#' Required only when the matching filter is enabled:
#' \itemize{
#'   \item `EG.Identified` -> `filter_by_identified`
#'   \item `F.ExcludedFromQuantification` -> `filter_by_excluded`
#'   \item `EG.Qvalue` and `PG.Qvalue` -> `filter_by_qvalue`
#' }
#' A missing required column raises an informative error naming the column.
#'
#' @export
#'
#' @examples
#' converted_data <- bigSpectronauttoMSstatsFormat(
#'   system.file("extdata", "spectronaut_input.csv", package = "MSstatsBig"),
#'   "output_file.csv",
#'   backend = "arrow")
#' converted_data <- dplyr::collect(converted_data)
#' head(converted_data)
#'
#' # Override Spectronaut's embedded Condition / BioReplicate with
#' # a custom annotation (e.g. for a paired design):
#' annot <- data.frame(Run = unique(converted_data[["Run"]]))
#' annot$BioReplicate <- seq_len(nrow(annot))
#' annot$Condition <- rep(c("ctrl", "treat"), length.out = nrow(annot))
#' overridden <- bigSpectronauttoMSstatsFormat(
#'   system.file("extdata", "spectronaut_input.csv", package = "MSstatsBig"),
#'   "output_file.csv",
#'   backend = "arrow",
#'   annotation = annot)
#' head(dplyr::collect(overridden))
#'
#' @return either arrow object or sparklyr table that can be optionally collected
#' into memory by using dplyr::collect function.
#'
bigSpectronauttoMSstatsFormat <-  function(input_file, output_file_name,
                                          backend,
                                          intensity = "F.NormalizedPeakArea",
                                          filter_by_excluded = FALSE,
                                          filter_by_identified = FALSE,
                                          filter_by_qvalue = FALSE,
                                          qvalue_cutoff = 0.01,
                                          max_feature_count = 100,
                                          filter_unique_peptides =  FALSE,
                                          aggregate_psms =  FALSE,
                                          filter_few_obs =  FALSE,
                                          remove_annotation =  FALSE,
                                          calculateAnomalyScores=FALSE,
                                          anomalyModelFeatures=c(),
                                          annotation = NULL,
                                          connection =  NULL,
                                          block_size = 16L * 1024L * 1024L) {
  reduced_file <- .prefixedPath("reduce_output_", output_file_name)
  reduceBigSpectronaut(input_file, reduced_file,
                       intensity, filter_by_excluded, filter_by_identified,
                       filter_by_qvalue, qvalue_cutoff,
                       calculateAnomalyScores, anomalyModelFeatures,
                       block_size = block_size)
  msstats_data <- MSstatsPreprocessBig(
    input_file = reduced_file,
    output_file_name = output_file_name,
    backend = backend,
    max_feature_count = max_feature_count,
    filter_unique_peptides = filter_unique_peptides,
    aggregate_psms = aggregate_psms,
    filter_few_obs = filter_few_obs,
    remove_annotation = remove_annotation,
    calculateAnomalyScores = calculateAnomalyScores,
    anomalyModelFeatures = anomalyModelFeatures,
    connection = connection)

  if (!is.null(annotation)) {
    msstats_data <- MSstatsAddAnnotationBig(msstats_data, annotation)
    if (backend == "arrow") {
      unlink(output_file_name, recursive = TRUE, force = TRUE)
      arrow::write_dataset(msstats_data, output_file_name, format = "csv")
    }
  }

  return(msstats_data)

}


#' Convert out-of-memory DIANN files to MSstats format.
#'
#' @inheritParams MSstatsPreprocessBig
#' @inheritParams MSstatsConvert::DIANNtoMSstatsFormat
#'
#' @export
#'
#' @return either arrow object or sparklyr table that can be optionally collected
#' into memory by using dplyr::collect function.
#'
bigDIANNtoMSstatsFormat <- function(input_file, 
                                    annotation = NULL,
                                    output_file_name,
                                    backend,
                                    MBR = TRUE,
                                    quantificationColumn = "FragmentQuantCorrected",
                                    global_qvalue_cutoff = 0.01,
                                    qvalue_cutoff = 0.01,
                                    pg_qvalue_cutoff = 0.01,
                                    max_feature_count = 100,
                                    filter_unique_peptides =  FALSE,
                                    aggregate_psms =  FALSE,
                                    filter_few_obs =  FALSE,
                                    remove_annotation =  FALSE,
                                    calculateAnomalyScores=FALSE, 
                                    anomalyModelFeatures=c(),
                                    connection =  NULL) {
  
  # Reduce and clean the DIANN report file in chunks
  reduced_file <- .prefixedPath("reduce_output_", output_file_name)
  reduceBigDIANN(input_file,
                 reduced_file,
                 MBR,
                 quantificationColumn,
                 global_qvalue_cutoff, qvalue_cutoff, pg_qvalue_cutoff,
                 calculateAnomalyScores, anomalyModelFeatures,
                 annotation)

  reduced <- arrow::open_dataset(reduced_file, format = "csv")

  # Identify columns where Arrow inferred 'null' type (all values NA)
  null_cols <- names(reduced$schema)[
    vapply(reduced$schema$fields, function(f) f$type$ToString() == "null", logical(1))
  ]

  if (length(null_cols) > 0) {
    # Drop null-typed columns using a lazy select (no data loaded into memory)
    reduced <- dplyr::select(reduced, -dplyr::all_of(null_cols))

    # Write back using Arrow's streaming writer — stays out-of-memory.
    # write_dataset creates a directory, but open_dataset can read
    # directories just as easily as single files.
    cleaned_file <- .prefixedPath("cleaned_", output_file_name)
    arrow::write_dataset(reduced, cleaned_file, format = "csv")
    reduced_file <- cleaned_file
  }

  # Preprocess the cleaned data (feature selection, etc.)
  msstats_data <- MSstatsPreprocessBig(
    input_file = reduced_file,
    output_file_name = output_file_name,
    backend = backend,
    max_feature_count = max_feature_count,
    filter_unique_peptides = filter_unique_peptides,
    aggregate_psms = aggregate_psms,
    filter_few_obs = filter_few_obs,
    remove_annotation = remove_annotation,
    calculateAnomalyScores = calculateAnomalyScores,
    anomalyModelFeatures = anomalyModelFeatures,
    connection = connection)

  # Merge annotation with the preprocessed data and persist the merge so
  # callers reopening output_file_name see Condition/BioReplicate. The arrow
  # rewrite stays lazy — the underlying source is reduced_file, not
  # output_file_name, so we can safely overwrite the directory we just wrote.
  if (!is.null(annotation)) {
    msstats_data <- MSstatsAddAnnotationBig(msstats_data, annotation)
    if (backend == "arrow") {
      unlink(output_file_name, recursive = TRUE, force = TRUE)
      arrow::write_dataset(msstats_data, output_file_name, format = "csv")
    }
  }
  return(msstats_data)
}


#' Merge annotation to output of MSstatsPreprocessBig
#'
#' @param input output of MSstatsPreprocessBig
#' @param annotation run annotation
#'
#' @export
#'
#' @examples
#' converted_data <- bigFragPipetoMSstatsFormat(
#'   system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
#'   "output_file.csv",
#'   backend = "arrow")
#' converted_data <- dplyr::collect(converted_data)
#' head(converted_data)
#' # Change annotation as an example:
#' converted_data$Condition <- NULL
#' converted_data$BioReplicate <- NULL
#' annot <- data.frame(Run = unique(converted_data[["Run"]]))
#' annot$BioReplicate <- rep(1:53, times = 2)
#' annot$Condition <- rep(1:2, each = 53)
#' head(MSstatsAddAnnotationBig(converted_data, annot))
#'
#' @importFrom MSstats dataProcess groupComparison
#' @importFrom utils head sessionInfo
#'
#' @return table of `input` and `annotation` merged by Run column.
#'
MSstatsAddAnnotationBig <- function(input, annotation) {
  join_keys <- "Run"
  
  # Use tbl_vars which works reliably on both Arrow
  # datasets, arrow_dplyr_query objects, and data frames
  input_cols <- dplyr::tbl_vars(input)
  
  overlap_cols <- setdiff(
    intersect(input_cols, colnames(annotation)),
    join_keys
  )
  if (length(overlap_cols) > 0) {
    input <- dplyr::select(input, -dplyr::all_of(overlap_cols))
  }
  
  dplyr::inner_join(input, annotation, by = join_keys)
}

library(testthat)
library(mockery)

context("General converter functions")

test_that("MSstatsAddAnnotationBig adds annotation correctly", {
  input_data <- data.frame(
    Run = c("Run1", "Run2", "Run3"),
    Intensity = c(100, 200, 300)
  )

  annotation_data <- data.frame(
    Run = c("Run1", "Run2", "Run3"),
    Condition = c("A", "A", "B"),
    BioReplicate = c(1, 2, 1)
  )

  expected_output <- data.frame(
    Run = c("Run1", "Run2", "Run3"),
    Intensity = c(100, 200, 300),
    Condition = c("A", "A", "B"),
    BioReplicate = c(1, 2, 1)
  )

  result <- MSstatsAddAnnotationBig(input_data, annotation_data)

  expect_equal(result, expected_output)
})

test_that("MSstatsPreprocessBig performs feature selection correctly", {
  input_file <- tempfile(fileext = ".csv")
  output_file <- "preprocess_output.csv"

  # P1 has 3 features (frag1, frag2, frag3). frag3 has the highest avg intensity.
  # P2 has 2 features (fragA, fragB). fragB has the highest avg intensity.
  msstats_data <- rbind(
    data.frame(ProteinName = "P1", PeptideSequence = "PEPTIDE", PrecursorCharge = 2, FragmentIon = rep(c("frag1", "frag2", "frag3"), each = 2), ProductCharge = 1, IsotopeLabelType = "L", Condition = "A", BioReplicate = rep(1:2, 3), Run = rep(c("run1", "run2"), 3), Intensity = c(1000, 1100, 500, 550, 2000, 2100)),
    data.frame(ProteinName = "P2", PeptideSequence = "PEPTIDE2", PrecursorCharge = 3, FragmentIon = rep(c("fragA", "fragB"), each = 2), ProductCharge = 1, IsotopeLabelType = "L", Condition = "B", BioReplicate = rep(1:2, 2), Run = rep(c("run1", "run2"), 2), Intensity = c(100, 150, 800, 850))
  )
  readr::write_csv(msstats_data, input_file)

  processed <- MSstatsPreprocessBig(input_file, output_file, backend = "arrow",
                                    max_feature_count = 1)
  result <- dplyr::collect(processed)

  # For P1, frag3 should be selected. For P2, fragB should be selected.
  expect_equal(nrow(result), 4)

  p1_result <- result[result$ProteinName == "P1", ]
  expect_equal(nrow(p1_result), 2)
  expect_true(all(p1_result$FragmentIon == "frag3"))

  p2_result <- result[result$ProteinName == "P2", ]
  expect_equal(nrow(p2_result), 2)
  expect_true(all(p2_result$FragmentIon == "fragB"))

  # Cleanup — output may be a directory when backend = "arrow"
  unlink(input_file, force = TRUE)
  unlink(output_file, recursive = TRUE, force = TRUE)
})

test_that("bigSpectronauttoMSstatsFormat works correctly", {
  # Mock reduceBigSpectronaut as its source is not provided
  mock_reduce <- mock(NULL)

  stub(bigSpectronauttoMSstatsFormat, "reduceBigSpectronaut", function(input_file, output_path, ...) {
    msstats_data <- data.frame(
      ProteinName = "P1", PeptideSequence = "PEPTIDE", PrecursorCharge = 2,
      FragmentIon = rep(c("frag1", "frag2"), each = 2), ProductCharge = 1,
      IsotopeLabelType = "L", Condition = "A", BioReplicate = rep(1:2, 2),
      Run = rep(c("run1", "run2"), 2), Intensity = c(1000, 1100, 2000, 2100) # frag2 is higher
    )
    readr::write_csv(msstats_data, output_path)
  })

  input_file <- "dummy_spectro_input.csv"
  output_file <- "spectro_output.csv"

  processed <- bigSpectronauttoMSstatsFormat(
    input_file = input_file,
    output_file_name = output_file,
    backend = "arrow",
    max_feature_count = 1
  )
  result <- dplyr::collect(processed)

  # The mock reduce function creates a file with 2 features for P1.
  # max_feature_count = 1 should select frag2.
  expect_equal(nrow(result), 2)
  expect_true(all(result$FragmentIon == "frag2"))

  # Cleanup — outputs may be directories when backend = "arrow"
  unlink(output_file, recursive = TRUE, force = TRUE)
  unlink(paste0("reduce_output_", output_file), recursive = TRUE, force = TRUE)
})

make_spectronaut_input <- function(n = 1L, ...) {
  base <- data.frame(
    R.FileName = rep("run1", n),
    R.Condition = "A",
    R.Replicate = 1L,
    PG.ProteinAccessions = "P1",
    EG.ModifiedSequence = paste0("PEP", seq_len(n)),
    FG.LabeledSequence = paste0("PEP", seq_len(n)),
    FG.Charge = 2L,
    F.FrgIon = "y1",
    F.Charge = 1L,
    EG.Identified = "True",
    F.ExcludedFromQuantification = "False",
    F.FrgLossType = "noloss",
    PG.Qvalue = 0.001,
    EG.Qvalue = 0.001,
    F.NormalizedPeakArea = seq_len(n) * 100,
    stringsAsFactors = FALSE
  )
  overrides <- list(...)
  for (col in names(overrides)) base[[col]] <- overrides[[col]]
  base
}

test_that("cleanSpectronautChunk produces the expected MSstats schema", {
  input <- make_spectronaut_input(n = 1L)
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L)
  result <- readr::read_csv(output_file, show_col_types = FALSE)

  expected_cols <- c("ProteinName", "PeptideSequence", "PrecursorCharge", "FragmentIon",
                     "ProductCharge", "IsotopeLabelType", "Run", "BioReplicate", "Condition",
                     "Intensity")
  expect_setequal(colnames(result), expected_cols)
  expect_equal(nrow(result), 1L)
  expect_equal(result$Intensity, 100)
  expect_equal(result$IsotopeLabelType, "L")
})

test_that("cleanSpectronautChunk filter_by_excluded sets Intensity to NA on excluded rows", {
  input <- make_spectronaut_input(
    n = 2L,
    F.ExcludedFromQuantification = c("True", "False"),
    F.NormalizedPeakArea = c(100, 200)
  )
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L,
                                     filter_by_excluded = TRUE,
                                     filter_by_qvalue = FALSE)
  result <- readr::read_csv(output_file, show_col_types = FALSE)
  result <- result[order(result$PeptideSequence), ]

  expect_true(is.na(result$Intensity[1]))
  expect_equal(result$Intensity[2], 200)
})

test_that("cleanSpectronautChunk filter_by_identified sets Intensity to NA on unidentified rows", {
  input <- make_spectronaut_input(
    n = 2L,
    EG.Identified = c("True", "False"),
    F.NormalizedPeakArea = c(100, 200)
  )
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L,
                                     filter_by_identified = TRUE,
                                     filter_by_qvalue = FALSE)
  result <- readr::read_csv(output_file, show_col_types = FALSE)
  result <- result[order(result$PeptideSequence), ]

  expect_equal(result$Intensity[1], 100)
  expect_true(is.na(result$Intensity[2]))
})

test_that("cleanSpectronautChunk filter_by_qvalue NA-aware semantics match dplyr::if_else", {
  input <- make_spectronaut_input(
    n = 3L,
    EG.Qvalue = c(0.001, 0.5, NA_real_),
    F.NormalizedPeakArea = c(100, 200, 300)
  )
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L,
                                     filter_by_qvalue = TRUE,
                                     qvalue_cutoff = 0.01)
  result <- readr::read_csv(output_file, show_col_types = FALSE)
  result <- result[order(result$PeptideSequence), ]

  # PEP1: EGQvalue below cutoff -> kept
  expect_equal(result$Intensity[1], 100)
  # PEP2: EGQvalue above cutoff -> NA
  expect_true(is.na(result$Intensity[2]))
  # PEP3: EGQvalue NA -> NA (preserves dplyr::if_else semantics)
  expect_true(is.na(result$Intensity[3]))
})

test_that("cleanSpectronautChunk drops rows where FFrgLossType != noloss", {
  input <- make_spectronaut_input(
    n = 3L,
    F.FrgLossType = c("noloss", "H2O", "noloss")
  )
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L,
                                     filter_by_qvalue = FALSE)
  result <- readr::read_csv(output_file, show_col_types = FALSE)

  expect_equal(nrow(result), 2L)
  expect_setequal(result$PeptideSequence, c("PEP1", "PEP3"))
})

test_that("cleanSpectronautChunk fails fast with a clear error when a filter column is absent", {
  output_file <- tempfile(fileext = ".csv")
  on.exit(unlink(output_file, force = TRUE), add = TRUE)

  # FFrgLossType filter always runs: dropping the source column must error,
  # naming the original Spectronaut column, not raise a cryptic data.table error.
  input <- make_spectronaut_input(n = 1L)
  input$F.FrgLossType <- NULL
  expect_error(
    MSstatsBig:::cleanSpectronautChunk(input, output_file, pos = 1L,
                                       filter_by_qvalue = FALSE),
    regexp = "F.FrgLossType")

  # q-value filter only errors when it is actually requested.
  input_noq <- make_spectronaut_input(n = 1L)
  input_noq$EG.Qvalue <- NULL
  expect_error(
    MSstatsBig:::cleanSpectronautChunk(input_noq, output_file, pos = 1L,
                                       filter_by_qvalue = TRUE),
    regexp = "EG.Qvalue")

  # ...and is tolerated when the q-value filter is switched off.
  expect_error(
    MSstatsBig:::cleanSpectronautChunk(make_spectronaut_input(n = 1L),
                                       output_file, pos = 1L,
                                       filter_by_qvalue = FALSE),
    NA)
})

test_that("reduceBigSpectronaut projects out columns cleanSpectronautChunk does not use", {
  # Real Spectronaut columns plus two junk columns that are not in needed_cols.
  input <- make_spectronaut_input(n = 2L)
  input$JunkA <- "zzz"
  input$JunkB <- 999L
  input_file <- tempfile(fileext = ".csv")
  output_file <- tempfile()
  on.exit({
    unlink(input_file, force = TRUE)
    unlink(output_file, recursive = TRUE, force = TRUE)
  }, add = TRUE)
  readr::write_csv(input, input_file)

  # Capture the columns of each batch actually handed to the cleaner. The
  # Scanner projection should have dropped JunkA/JunkB before this point.
  captured <- new.env(parent = emptyenv())
  captured$cols <- NULL
  stub(reduceBigSpectronaut, "cleanSpectronautChunk",
       function(input, ...) {
         captured$cols <- colnames(input)
         NULL
       })

  reduceBigSpectronaut(input_file, output_file)

  expect_false("JunkA" %in% captured$cols)
  expect_false("JunkB" %in% captured$cols)
  # Needed columns still make it through.
  expect_true("F.NormalizedPeakArea" %in% captured$cols)
  expect_true("PG.ProteinAccessions" %in% captured$cols)
})

test_that("reduceBigSpectronaut rejects invalid block_size values", {
  input_file <- tempfile(fileext = ".csv")
  writeLines("a,b\n1,2", input_file)
  output_file <- tempfile()
  on.exit({
    unlink(input_file, force = TRUE)
    unlink(output_file, recursive = TRUE, force = TRUE)
  }, add = TRUE)

  expect_error(reduceBigSpectronaut(input_file, output_file, block_size = -1L),
               regexp = "block_size")
  expect_error(reduceBigSpectronaut(input_file, output_file, block_size = 0L),
               regexp = "block_size")
  expect_error(reduceBigSpectronaut(input_file, output_file, block_size = NA_integer_),
               regexp = "block_size")
  expect_error(reduceBigSpectronaut(input_file, output_file, block_size = c(1L, 2L)),
               regexp = "block_size")
  expect_error(suppressWarnings(
    reduceBigSpectronaut(input_file, output_file, block_size = "16MB")
  ), regexp = "block_size")
})

test_that("bigSpectronauttoMSstatsFormat plumbs block_size through to reduceBigSpectronaut", {
  captured <- new.env(parent = emptyenv())
  captured$block_size <- NULL

  spy_reduce <- function(input_file, output_path, intensity, filter_by_excluded,
                        filter_by_identified, filter_by_qvalue, qvalue_cutoff,
                        calculateAnomalyScores, anomalyModelFeatures,
                        block_size = 16L * 1024L * 1024L) {
    captured$block_size <- block_size
    msstats_data <- data.frame(
      ProteinName = "P1", PeptideSequence = "PEPTIDE", PrecursorCharge = 2,
      FragmentIon = "frag1", ProductCharge = 1,
      IsotopeLabelType = "L", Condition = "A", BioReplicate = 1,
      Run = "run1", Intensity = 100
    )
    readr::write_csv(msstats_data, output_path)
  }

  input_file <- "dummy_spectro_input.csv"

  # Default forwards 16 MiB.
  stub(bigSpectronauttoMSstatsFormat, "reduceBigSpectronaut", spy_reduce)
  output_file_default <- tempfile(fileext = ".csv")
  on.exit({
    unlink(output_file_default, recursive = TRUE, force = TRUE)
    unlink(paste0("reduce_output_", basename(output_file_default)),
           recursive = TRUE, force = TRUE)
  }, add = TRUE)
  bigSpectronauttoMSstatsFormat(
    input_file = input_file,
    output_file_name = output_file_default,
    backend = "arrow"
  )
  expect_identical(captured$block_size, 16L * 1024L * 1024L)

  # Override forwards the user's value.
  output_file_override <- tempfile(fileext = ".csv")
  on.exit({
    unlink(output_file_override, recursive = TRUE, force = TRUE)
    unlink(paste0("reduce_output_", basename(output_file_override)),
           recursive = TRUE, force = TRUE)
  }, add = TRUE)
  bigSpectronauttoMSstatsFormat(
    input_file = input_file,
    output_file_name = output_file_override,
    backend = "arrow",
    block_size = 8L * 1024L * 1024L
  )
  expect_identical(captured$block_size, 8L * 1024L * 1024L)
})

---
title: "MSstatsBig Workflow"
author: "Devon Kohler (<kohler.d@northeastern.edu>)"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{MSstatsBig Workflow}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## MSstatsBig Package Description

`MSstatsBig` is designed to overcome challenges when analyzing very large mass 
spectrometry (MS)-based proteomics experiments. These experiments are generally
(but not always) acquired with DIA and include a large number of MS runs. 
`MSstatsBig` leverages software that can work on datasets without loading them
into memory. This avoids a major problem where a dataset cannot be loaded into a
standard computers RAM.

`MSstatsBig` includes functions which are designed to replace the converters 
included in the `MSstats` package. The goal of these converters is to perform 
filtering on the PSM files of identified and quantified data to remove data that
is not required for differential analysis. Once this data is filtered down it 
should be able to be loaded into your computer's memory. After the converters 
are run, the standard `MSstats` workflow can be followed.

`MSstatsBig` currently includes converters for Spectronaut and FragPipe. Beyond 
these converters, users can manually use the underlying functions by putting 
their data into `MSstats` format and running the underlying 
`MSstatsPreprocessBig` function.
This way, either by using native export format of signal processing tools or by 
converting raw data chunk by chunk 
(for example with the readr::read_delim_chunked function), `MSstatsBig` can be 
used with other popular tools such as DIA-NN.

``` {r load_package}
library(MSstatsBig)
```

## Dataset description

The dataset included in this package is a small subset of a work by Clark et 
al. [1]. It is a large DIA dataset that includes over 100 runs. The experimental
data was identified and quantified by FragPipe and the included dataset is the 
`msstats.csv` output for FragPipe.

```{r example_data}
head(read.csv(system.file("extdata", "fgexample.csv", package = "MSstatsBig")))
```

## Run MSstatsBig converter

First we run the `MSstatsBig` converter. The converter will save the dataset to 
a place on your computer, and will return an arrow object. Once then, you can 
read the data from text file or load the arrow data.frame into memory by using
the `dplyr::collect` function. The "collected" data can then be treated as a 
standard R `data.frame`.

``` {r conv_example}
setwd(tempdir())

converted_data = bigFragPipetoMSstatsFormat(
  system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend="arrow",
  max_feature_count = 20)

# The returned arrow object needs to be collected for the remaining workflow
converted_data = as.data.frame(dplyr::collect(converted_data))
```

##  Remaining workflow

Once the converter is run the standard `MSstats` workflow can be followed:

  * `dataProcess` function summarizes peptide-level data into protein-level estimates.
  * `groupComparison` function fits a linear model to protein-level data and
  performs statistical inference for comparisons of selected groups.
  
Details of the MSstats workflow can be found in [2].

``` {r msstats_workflow}

library(MSstats)

# converted_data = read.csv("output_file.csv")
summarized_data = dataProcess(converted_data,
                              use_log_file = FALSE)

# Build contrast matrix
comparison = matrix(c(-1, 1),
    nrow=1, byrow=TRUE)
row.names(comparison) <- c("T-NAT")
colnames(comparison) <- c("NAT", "T")

model_results = groupComparison(contrast.matrix = comparison, 
                                data = summarized_data,
                                use_log_file = FALSE)


```

## Session info

```{r }
sessionInfo()
```

## References

1. D. J. Clark, S. M. Dhanasekaran, F. Petralia, P. Wang and H. Zhang, 
"Integrated Proteogenomic Characterization of Clear Cell Renal Cell Carcinoma," 
Cell, vol. 179, pp. 964-983, 2019. 

2. D. Kohler et al., "MSstats Version 4.0: Statistical Analyses of Quantitative 
Mass Spectrometry-Based Proteomic Experiments with Chromatography-Based 
Quantification at Scale", J. Proteome Res. 22, 5, pp. 1466–1482, 
J. Proteome Res. 2023, 22, 5, 1466–1482

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{MSstatsPreprocessBig}
\alias{MSstatsPreprocessBig}
\title{General converter for larger-than-memory csv files in MSstats format 10-column format}
\usage{
MSstatsPreprocessBig(
  input_file,
  output_file_name,
  backend,
  max_feature_count = 100,
  filter_unique_peptides = FALSE,
  aggregate_psms = FALSE,
  filter_few_obs = FALSE,
  remove_annotation = FALSE,
  calculateAnomalyScores = FALSE,
  anomalyModelFeatures = c(),
  connection = NULL
)
}
\arguments{
\item{input_file}{name of the input text file in 10-column MSstats format.}

\item{output_file_name}{name of an output file which will be saved after pre-processing}

\item{backend}{"arrow" or "sparklyr". Option "sparklyr" requires a spark installation
and connection to spark instance provided in the `connection` parameter.}

\item{max_feature_count}{maximum number of features per protein. Features will
be selected based on highest average intensity.}

\item{filter_unique_peptides}{If TRUE, shared peptides will be removed.
Please refer to the `Details` section for additional information.}

\item{aggregate_psms}{If TRUE, multiple measurements per PSM in a Run will
be aggregated (by taking maximum value). Please refer to the `Details` section for additional information.}

\item{filter_few_obs}{If TRUE, feature with less than 3 observations across runs will be removed.
Please refer to the `Details` section for additional information.}

\item{remove_annotation}{If TRUE, columns BioReplicate and Condition will be removed
to reduce output file size. These will need to be added manually later before
using dataProcess function. Only applicable to sparklyr backend.}

\item{calculateAnomalyScores}{If TRUE, will carry anomaly model features through pipeline}

\item{anomalyModelFeatures}{Character vector of column names to be carried through the pipeline}

\item{connection}{Connection to a spark instance created with the
`spark_connect` function from `sparklyr` package.}
}
\value{
either arrow object or sparklyr table that can be optionally collected
into memory by using dplyr::collect function.
}
\description{
General converter for larger-than-memory csv files in MSstats format 10-column format
}
\details{
Filtering and aggregation may be very time consuming and the ability
to perform them in a given R session depends on available memory, settings of
external packages, etc. Hence, all value of related parameters (`filter_unique_peptides`,
`aggregate_psms`, `filter_few_obs`) are set to FALSE by default and only feature
selection is performed, which saves both computation time and memory.
Appropriately configured spark backend provides the most consistent way to
perform these operations.
}
\examples{
converted_data <- bigFragPipetoMSstatsFormat(
  system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
  "tencol_format.csv",
  backend="arrow")
procd <- MSstatsPreprocessBig("tencol_format.csv", "proc_out.csv", backend = "arrow")
head(dplyr::collect(procd))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_DIANN.R
\name{reduceBigDIANN}
\alias{reduceBigDIANN}
\title{Read and clean a large DIANN file in chunks}
\usage{
reduceBigDIANN(
  input_file,
  output_path,
  MBR = TRUE,
  quantificationColumn = "FragmentQuantCorrected",
  global_qvalue_cutoff = 0.01,
  qvalue_cutoff = 0.01,
  pg_qvalue_cutoff = 0.01,
  calculateAnomalyScores = FALSE,
  anomalyModelFeatures = c(),
  annotation = NULL
)
}
\arguments{
\item{input_file}{Path to the input DIANN file}

\item{output_path}{Path to the output CSV file}

\item{MBR}{Boolean, whether MBR was used}

\item{quantificationColumn}{Name of the column containing intensity values}

\item{global_qvalue_cutoff}{Global Q-value cutoff}

\item{qvalue_cutoff}{Q-value cutoff}

\item{pg_qvalue_cutoff}{Protein group Q-value cutoff}

\item{calculateAnomalyScores}{Boolean for MSstats+ Model}

\item{anomalyModelFeatures}{Character vector of features to use for MSstats+ Model}

\item{annotation}{Annotation file or data frame}
}
\value{
NULL. Writes to file.
}
\description{
Read and clean a large DIANN file in chunks
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.writeChunkToFile}
\alias{.writeChunkToFile}
\title{Write chunk to file}
\usage{
.writeChunkToFile(input, output_path, pos)
}
\arguments{
\item{input}{Data frame}

\item{output_path}{Path to output file}

\item{pos}{Chunk position}
}
\description{
Write chunk to file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{bigSpectronauttoMSstatsFormat}
\alias{bigSpectronauttoMSstatsFormat}
\title{Convert out-of-memory Spectronaut files to MSstats format.}
\usage{
bigSpectronauttoMSstatsFormat(
  input_file,
  output_file_name,
  backend,
  intensity = "F.NormalizedPeakArea",
  filter_by_excluded = FALSE,
  filter_by_identified = FALSE,
  filter_by_qvalue = FALSE,
  qvalue_cutoff = 0.01,
  max_feature_count = 100,
  filter_unique_peptides = FALSE,
  aggregate_psms = FALSE,
  filter_few_obs = FALSE,
  remove_annotation = FALSE,
  calculateAnomalyScores = FALSE,
  anomalyModelFeatures = c(),
  annotation = NULL,
  connection = NULL,
  block_size = 16L * 1024L * 1024L
)
}
\arguments{
\item{input_file}{name of the input text file in 10-column MSstats format.}

\item{output_file_name}{name of an output file which will be saved after pre-processing}

\item{backend}{"arrow" or "sparklyr". Option "sparklyr" requires a spark installation
and connection to spark instance provided in the `connection` parameter.}

\item{intensity}{Name of the intensity column to be used in Spectronaut}

\item{filter_by_excluded}{if TRUE, will filter by the `F.ExcludedFromQuantification` column.}

\item{filter_by_identified}{if TRUE, will filter by the `EG.Identified` column.}

\item{filter_by_qvalue}{if TRUE, will filter by EG.Qvalue and PG.Qvalue columns.}

\item{qvalue_cutoff}{cutoff which will be used for q-value filtering.}

\item{max_feature_count}{maximum number of features per protein. Features will
be selected based on highest average intensity.}

\item{filter_unique_peptides}{If TRUE, shared peptides will be removed.
Please refer to the `Details` section for additional information.}

\item{aggregate_psms}{If TRUE, multiple measurements per PSM in a Run will
be aggregated (by taking maximum value). Please refer to the `Details` section for additional information.}

\item{filter_few_obs}{If TRUE, feature with less than 3 observations across runs will be removed.
Please refer to the `Details` section for additional information.}

\item{remove_annotation}{If TRUE, columns BioReplicate and Condition will be removed
to reduce output file size. These will need to be added manually later before
using dataProcess function. Only applicable to sparklyr backend.}

\item{calculateAnomalyScores}{If TRUE, will carry anomaly model features through pipeline}

\item{anomalyModelFeatures}{Character vector of column names to be carried through the pipeline}

\item{annotation}{Optional data.frame with columns `Run`,
`BioReplicate`, `Condition` (plus any additional annotation
columns). If supplied, the converter merges it onto the output
on `Run` and overrides any `Condition` / `BioReplicate` values
that came from Spectronaut's `R.Condition` / `R.Replicate`
columns. Required when the experimental design cannot be
expressed in Spectronaut's own annotation — most notably paired
designs, where `BioReplicate` must encode the pairing
structure rather than per-sample IDs.}

\item{connection}{Connection to a spark instance created with the
`spark_connect` function from `sparklyr` package.}

\item{block_size}{Arrow CSV reader block size in bytes: the amount of raw
CSV that is read and parsed at a time (roughly one record batch per block).
Defaults to 16 MiB (`16L * 1024L * 1024L`). Two things matter when tuning it:
\itemize{
  \item It must be larger than your single widest row. If you see
    `Invalid: straddling object straddles two block boundaries` on
    extra-wide exports, raise it (e.g. `64L * 1024L * 1024L`).
  \item Larger blocks reduce parsing overhead, but peak memory grows
    roughly linearly with `block_size`, so it is also the main lever on
    out-of-memory risk. Gains taper off quickly because per-batch time is
    dominated by row cleaning, not by block size.
}
To speed up processing without risking a crash, double the value
(32 -> 64 -> 128 MiB) while watching memory use, and stop once throughput
stops improving. Keep it comfortably below free RAM -- and below free RAM
divided by the number of jobs if you run several conversions in parallel.}
}
\value{
either arrow object or sparklyr table that can be optionally collected
into memory by using dplyr::collect function.
}
\description{
Convert out-of-memory Spectronaut files to MSstats format.
}
\details{
`bigSpectronauttoMSstatsFormat` consumes the following Spectronaut export
columns.

Always required:
\itemize{
  \item `R.FileName`, `R.Condition`, `R.Replicate` -> Run, Condition, BioReplicate
  \item `PG.ProteinAccessions` -> ProteinName
  \item `EG.ModifiedSequence` -> PeptideSequence
  \item `FG.Charge` -> PrecursorCharge; `F.FrgIon` -> FragmentIon; `F.Charge` -> ProductCharge
  \item the intensity column (`F.NormalizedPeakArea` by default; see `intensity`) -> Intensity
  \item `F.FrgLossType` -> used by the always-on `noloss` fragment filter
}


Required only when the matching filter is enabled:
\itemize{
  \item `EG.Identified` -> `filter_by_identified`
  \item `F.ExcludedFromQuantification` -> `filter_by_excluded`
  \item `EG.Qvalue` and `PG.Qvalue` -> `filter_by_qvalue`
}
A missing required column raises an informative error naming the column.
}
\examples{
converted_data <- bigSpectronauttoMSstatsFormat(
  system.file("extdata", "spectronaut_input.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend = "arrow")
converted_data <- dplyr::collect(converted_data)
head(converted_data)

# Override Spectronaut's embedded Condition / BioReplicate with
# a custom annotation (e.g. for a paired design):
annot <- data.frame(Run = unique(converted_data[["Run"]]))
annot$BioReplicate <- seq_len(nrow(annot))
annot$Condition <- rep(c("ctrl", "treat"), length.out = nrow(annot))
overridden <- bigSpectronauttoMSstatsFormat(
  system.file("extdata", "spectronaut_input.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend = "arrow",
  annotation = annot)
head(dplyr::collect(overridden))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.prefixedPath}
\alias{.prefixedPath}
\title{Build an intermediate output path by prefixing only the basename.}
\usage{
.prefixedPath(prefix, path)
}
\arguments{
\item{prefix}{Character scalar prepended to the basename.}

\item{path}{Output file path supplied by the caller.}
}
\value{
Character scalar.
}
\description{
Naive `paste0(prefix, output_file_name)` corrupts paths that contain a
directory (`subdir/out.csv` → `topN_subdir/out.csv`,
`/tmp/out.csv` → `topN_/tmp/out.csv`). Splitting via dirname/basename keeps
the directory component intact so intermediate files land beside the final
output.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_DIANN.R
\name{cleanDIANNChunk}
\alias{cleanDIANNChunk}
\title{Clean a single chunk of DIANN data}
\usage{
cleanDIANNChunk(
  input,
  output_path,
  MBR,
  quantificationColumn,
  pos,
  global_qvalue_cutoff = 0.01,
  qvalue_cutoff = 0.01,
  pg_qvalue_cutoff = 0.01,
  calculateAnomalyScores = FALSE,
  anomalyModelFeatures = c(),
  annotation = NULL
)
}
\arguments{
\item{input}{Data frame chunk}

\item{output_path}{Path to output file}

\item{MBR}{Boolean, whether MBR was used}

\item{quantificationColumn}{Name of intensity column}

\item{pos}{Chunk position (1 for first chunk, >1 for subsequent)}

\item{global_qvalue_cutoff}{Global Q-value cutoff}

\item{qvalue_cutoff}{Q-value cutoff}

\item{pg_qvalue_cutoff}{Protein group Q-value cutoff}

\item{calculateAnomalyScores}{Boolean for MSstats+ Model}

\item{anomalyModelFeatures}{Character vector of features to use for MSstats+ Model}

\item{annotation}{Annotation file or data frame}
}
\description{
Clean a single chunk of DIANN data
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{bigDIANNtoMSstatsFormat}
\alias{bigDIANNtoMSstatsFormat}
\title{Convert out-of-memory DIANN files to MSstats format.}
\usage{
bigDIANNtoMSstatsFormat(
  input_file,
  annotation = NULL,
  output_file_name,
  backend,
  MBR = TRUE,
  quantificationColumn = "FragmentQuantCorrected",
  global_qvalue_cutoff = 0.01,
  qvalue_cutoff = 0.01,
  pg_qvalue_cutoff = 0.01,
  max_feature_count = 100,
  filter_unique_peptides = FALSE,
  aggregate_psms = FALSE,
  filter_few_obs = FALSE,
  remove_annotation = FALSE,
  calculateAnomalyScores = FALSE,
  anomalyModelFeatures = c(),
  connection = NULL
)
}
\arguments{
\item{input_file}{name of the input text file in 10-column MSstats format.}

\item{annotation}{name of 'annotation.txt' data which includes Condition, BioReplicate, Run.}

\item{output_file_name}{name of an output file which will be saved after pre-processing}

\item{backend}{"arrow" or "sparklyr". Option "sparklyr" requires a spark installation
and connection to spark instance provided in the `connection` parameter.}

\item{MBR}{True if analysis was done with match between runs}

\item{quantificationColumn}{Use 'FragmentQuantCorrected'(default) column for quantified intensities for DIANN 1.8.x.
Use 'FragmentQuantRaw' for quantified intensities for DIANN 1.9.x.
Use 'auto' for quantified intensities for DIANN 2.x where each fragment intensity is a separate column, e.g. Fr0Quantity.}

\item{global_qvalue_cutoff}{The qvalue cutoff for the Q.Value column, i.e.
the run-specific precursor q-value. Default is 0.01.}

\item{qvalue_cutoff}{If MBR is false, the qvalue cutoff for the Global.Q.Value
column, i.e. global precursor q-value.  If MBR is true, the qvalue cutoff for the
Lib.Q.Value column, i.e. the q-value for the library created after the first MBR pass.
Default is 0.01.}

\item{pg_qvalue_cutoff}{If MBR is false, the qvalue cutoff for the Global.PG.Q.Value
column, i.e. the global q-value for the protein group.  If MBR is true, the
qvalue cutoff for the Lib.PG.Q.Value column, i.e. the protein group q-value for
the library created after the first MBR pass. Default is 0.01.}

\item{max_feature_count}{maximum number of features per protein. Features will
be selected based on highest average intensity.}

\item{filter_unique_peptides}{If TRUE, shared peptides will be removed.
Please refer to the `Details` section for additional information.}

\item{aggregate_psms}{If TRUE, multiple measurements per PSM in a Run will
be aggregated (by taking maximum value). Please refer to the `Details` section for additional information.}

\item{filter_few_obs}{If TRUE, feature with less than 3 observations across runs will be removed.
Please refer to the `Details` section for additional information.}

\item{remove_annotation}{If TRUE, columns BioReplicate and Condition will be removed
to reduce output file size. These will need to be added manually later before
using dataProcess function. Only applicable to sparklyr backend.}

\item{calculateAnomalyScores}{If TRUE, will carry anomaly model features through pipeline}

\item{anomalyModelFeatures}{Character vector of column names to be carried through the pipeline}

\item{connection}{Connection to a spark instance created with the
`spark_connect` function from `sparklyr` package.}
}
\value{
either arrow object or sparklyr table that can be optionally collected
into memory by using dplyr::collect function.
}
\description{
Convert out-of-memory DIANN files to MSstats format.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{bigFragPipetoMSstatsFormat}
\alias{bigFragPipetoMSstatsFormat}
\title{Convert out-of-memory FragPipe files to MSstats format.}
\usage{
bigFragPipetoMSstatsFormat(
  input_file,
  output_file_name,
  backend,
  max_feature_count = 100,
  filter_unique_peptides = FALSE,
  aggregate_psms = FALSE,
  filter_few_obs = FALSE,
  remove_annotation = FALSE,
  connection = NULL
)
}
\arguments{
\item{input_file}{name of the input text file in 10-column MSstats format.}

\item{output_file_name}{name of an output file which will be saved after pre-processing}

\item{backend}{"arrow" or "sparklyr". Option "sparklyr" requires a spark installation
and connection to spark instance provided in the `connection` parameter.}

\item{max_feature_count}{maximum number of features per protein. Features will
be selected based on highest average intensity.}

\item{filter_unique_peptides}{If TRUE, shared peptides will be removed.
Please refer to the `Details` section for additional information.}

\item{aggregate_psms}{If TRUE, multiple measurements per PSM in a Run will
be aggregated (by taking maximum value). Please refer to the `Details` section for additional information.}

\item{filter_few_obs}{If TRUE, feature with less than 3 observations across runs will be removed.
Please refer to the `Details` section for additional information.}

\item{remove_annotation}{If TRUE, columns BioReplicate and Condition will be removed
to reduce output file size. These will need to be added manually later before
using dataProcess function. Only applicable to sparklyr backend.}

\item{connection}{Connection to a spark instance created with the
`spark_connect` function from `sparklyr` package.}
}
\value{
either arrow object or sparklyr table that can be optionally collected
into memory by using dplyr::collect function.
}
\description{
Convert out-of-memory FragPipe files to MSstats format.
}
\examples{
converted_data <- bigFragPipetoMSstatsFormat(
  system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend = "arrow")
converted_data <- dplyr::collect(converted_data)
head(converted_data)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/converters.R
\name{MSstatsAddAnnotationBig}
\alias{MSstatsAddAnnotationBig}
\title{Merge annotation to output of MSstatsPreprocessBig}
\usage{
MSstatsAddAnnotationBig(input, annotation)
}
\arguments{
\item{input}{output of MSstatsPreprocessBig}

\item{annotation}{run annotation}
}
\value{
table of `input` and `annotation` merged by Run column.
}
\description{
Merge annotation to output of MSstatsPreprocessBig
}
\examples{
converted_data <- bigFragPipetoMSstatsFormat(
  system.file("extdata", "fgexample.csv", package = "MSstatsBig"),
  "output_file.csv",
  backend = "arrow")
converted_data <- dplyr::collect(converted_data)
head(converted_data)
# Change annotation as an example:
converted_data$Condition <- NULL
converted_data$BioReplicate <- NULL
annot <- data.frame(Run = unique(converted_data[["Run"]]))
annot$BioReplicate <- rep(1:53, times = 2)
annot$Condition <- rep(1:2, each = 53)
head(MSstatsAddAnnotationBig(converted_data, annot))

}
